import {
  PROPOSAL_COLOR,
  type ProposalStatusColor,
} from "$lib/constants/proposals.constants";
import { i18n } from "$lib/stores/i18n";
import type { ProposalsFiltersStore } from "$lib/stores/proposals.store";
import type { VoteRegistration } from "$lib/stores/vote-registration.store";
import type { Identity } from "@dfinity/agent";
import type {
  Ballot,
  ExecuteNnsFunction,
  NeuronId,
  NeuronInfo,
  Proposal,
  ProposalId,
  ProposalInfo,
  Tally,
} from "@dfinity/nns";
import {
  NnsFunction,
  ProposalRewardStatus,
  ProposalStatus,
  Topic,
  Vote,
} from "@dfinity/nns";
import { get } from "svelte/store";
import { nowInSeconds } from "./date.utils";
import { errorToString } from "./error.utils";
import { replacePlaceholders } from "./i18n.utils";
import { isDefined, keyOf, keyOfOptional, nonNullish } from "./utils";

export const lastProposalId = (
  proposalInfos: ProposalInfo[]
): ProposalId | undefined => {
  const { length, [length - 1]: last } = proposalInfos;
  return last?.id;
};

export const proposalFirstActionKey = (
  proposal: Proposal | undefined
): string | undefined => Object.keys(proposal?.action ?? {})[0];

export const proposalActionFields = (
  proposal: Proposal
): [string, unknown][] => {
  const key = proposalFirstActionKey(proposal);
  if (key === undefined) {
    return [];
  }
  return Object.entries(
    keyOfOptional({ obj: proposal.action, key }) ?? {}
  ).filter(([, value]) => {
    switch (typeof value) {
      case "object":
        return value && Object.keys(value).length > 0;
      case "string":
      case "bigint":
      case "boolean":
      case "number":
        return true;
    }
    return false;
  });
};

export const getNnsFunctionKey = (
  proposal: Proposal | undefined
): string | undefined => {
  const action = proposalFirstActionKey(proposal);

  if (action !== "ExecuteNnsFunction") {
    return undefined;
  }

  // 0 equals Unspecified
  const { nnsFunctionId }: ExecuteNnsFunction = keyOfOptional({
    obj: proposal?.action,
    key: action,
  }) ?? {
    nnsFunctionId: 0,
  };

  return NnsFunction[nnsFunctionId];
};

/**
 * Hide proposal that don't match filters
 *
 * And check whether we hide it because the user has already voted on it and doesn't want to see them.
 */
export const hideProposal = ({
  proposalInfo,
  filters,
  neurons,
  identity,
}: {
  proposalInfo: ProposalInfo;
  filters: ProposalsFiltersStore;
  neurons: NeuronInfo[];
  identity: Identity | undefined | null;
}): boolean =>
  !matchFilters({ proposalInfo, filters }) ||
  (nonNullish(identity) &&
    !identity.getPrincipal().isAnonymous() &&
    isExcludedVotedProposal({ proposalInfo, filters, neurons }));

/**
 * Does the proposal returned by the backend really matches the filter selected by the user?
 */
const matchFilters = ({
  proposalInfo,
  filters,
}: {
  proposalInfo: ProposalInfo;
  filters: ProposalsFiltersStore;
}): boolean => {
  const { topics, rewards, status } = filters;

  const {
    topic: proposalTopic,
    status: proposalStatus,
    rewardStatus,
  } = proposalInfo;

  return (
    topics.includes(proposalTopic) &&
    rewards.includes(rewardStatus) &&
    status.includes(proposalStatus)
  );
};

/**
 * Hide a proposal if checkbox "excludeVotedProposals" is selected and the proposal's voting period has ended or has no UNSPECIFIED ballots' vote.
 */
const isExcludedVotedProposal = ({
  proposalInfo,
  filters,
  neurons,
}: {
  proposalInfo: ProposalInfo;
  filters: ProposalsFiltersStore;
  neurons: NeuronInfo[];
}): boolean => {
  const { excludeVotedProposals } = filters;

  const { ballots } = proposalInfo;
  const belongsToValidNeuron = (id: NeuronId) =>
    neurons.find(({ neuronId }) => neuronId === id) !== undefined;
  const containsUnspecifiedBallot = (): boolean =>
    ballots.find(
      ({ vote, neuronId }) =>
        // TODO: This is temporary solution. Will be replaced with L2-507
        // ignore neuronIds in ballots that are not in the neuron list of the user.
        // Otherwise it is confusing that there are proposals in the filtered list that can't vote.
        belongsToValidNeuron(neuronId) && vote === Vote.Unspecified
    ) !== undefined;

  const isOpen: boolean =
    isProposalDeadlineInTheFuture(proposalInfo) && containsUnspecifiedBallot();

  return excludeVotedProposals && !isOpen;
};

/**
 * Do we have any proposals that match the filters to render or should we display the user that nothing was found?
 */
export const hasMatchingProposals = ({
  proposals,
  filters,
  neurons,
  identity,
}: {
  proposals: ProposalInfo[];
  filters: ProposalsFiltersStore;
  neurons: NeuronInfo[];
  identity: Identity | undefined | null;
}): boolean => {
  if (proposals.length === 0) {
    return false;
  }

  return (
    proposals.find(
      (proposalInfo: ProposalInfo) =>
        !hideProposal({ proposalInfo, filters, neurons, identity })
    ) !== undefined
  );
};

export const getVotingBallot = ({
  neuronId,
  proposalInfo: { ballots },
}: {
  neuronId: bigint;
  proposalInfo: ProposalInfo;
}): Ballot | undefined =>
  ballots.find((ballot) => ballot.neuronId === neuronId);

// We first check the voting power of the ballot from the proposal. Which is the voting power that was used to vote.
// In the edge case that the proposal voting power is not present, then we show the neuron voting power.
export const getVotingPower = ({
  neuron: { neuronId, votingPower },
  proposal,
}: {
  neuron: NeuronInfo;
  proposal: ProposalInfo;
}): bigint =>
  getVotingBallot({
    neuronId,
    proposalInfo: proposal,
  })?.votingPower ?? votingPower;

export const selectedNeuronsVotingPower = ({
  neurons,
  selectedIds,
  proposal,
}: {
  neurons: NeuronInfo[];
  selectedIds: NeuronId[];
  proposal: ProposalInfo;
}): bigint =>
  neurons
    .filter(({ neuronId }) => selectedIds.includes(neuronId))
    .map((neuron) => getVotingPower({ neuron, proposal }))
    .reduce((sum, votingPower) => sum + votingPower, BigInt(0));

/**
 * Generate new selected neuron id list after new neurons response w/o spoiling the previously done user selection
 */
export const preserveNeuronSelectionAfterUpdate = ({
  selectedIds,
  neurons,
  updatedNeurons,
}: {
  selectedIds: NeuronId[];
  neurons: NeuronInfo[];
  updatedNeurons: NeuronInfo[];
}): NeuronId[] => {
  const newIds = new Set(updatedNeurons.map(({ neuronId }) => neuronId));
  const oldIds = new Set(neurons.map(({ neuronId }) => neuronId));
  const preservedSelection = selectedIds.filter((id) => newIds.has(id));
  const newNeuronsSelection = Array.from(newIds).filter(
    (id) => oldIds.has(id) === false
  );
  return [...preservedSelection, ...newNeuronsSelection];
};

export const proposalIdSet = (proposals: ProposalInfo[]): Set<ProposalId> =>
  new Set(proposals.map(({ id }) => id).filter(isDefined));

/**
 * Compares proposals by "id"
 */
export const concatenateUniqueProposals = ({
  oldProposals,
  newProposals,
}: {
  oldProposals: ProposalInfo[];
  newProposals: ProposalInfo[];
}): ProposalInfo[] => [
  ...oldProposals,
  ...excludeProposals({
    proposals: newProposals,
    exclusion: oldProposals,
  }),
];

/**
 * Compares proposals by "id" and replace those existing then append the remaining new proposals.
 */
export const replaceAndConcatenateProposals = ({
  oldProposals,
  newProposals,
}: {
  oldProposals: ProposalInfo[];
  newProposals: ProposalInfo[];
}): ProposalInfo[] => {
  const updatedProposals = oldProposals.map(
    (stateProposal) =>
      newProposals.find(({ id }) => stateProposal.id === id) ?? stateProposal
  );

  const newStateProposalsIds: (ProposalId | undefined)[] = updatedProposals.map(
    ({ id }) => id
  );
  const brandNewProposals: ProposalInfo[] = newProposals.filter(
    ({ id }) => !newStateProposalsIds.includes(id)
  );

  return concatenateUniqueProposals({
    oldProposals: updatedProposals,
    newProposals: brandNewProposals,
  });
};

/**
 * Compares proposals by "id"
 */
export const replaceProposals = ({
  oldProposals,
  newProposals,
}: {
  oldProposals: ProposalInfo[];
  newProposals: ProposalInfo[];
}): ProposalInfo[] =>
  oldProposals.map(
    (oldProposal) =>
      newProposals.find(({ id }) => id === oldProposal.id) ?? oldProposal
  );

/**
 * Compares 2 proposal lists by entries "id"
 */
export const proposalsHaveSameIds = ({
  proposalsA,
  proposalsB,
}: {
  proposalsA: ProposalInfo[];
  proposalsB: ProposalInfo[];
}): boolean =>
  proposalsA
    .map(({ id }) => id)
    .sort()
    .join() ===
  proposalsB
    .map(({ id }) => id)
    .sort()
    .join();

export const excludeProposals = ({
  proposals,
  exclusion,
}: {
  proposals: ProposalInfo[];
  exclusion: ProposalInfo[];
}): ProposalInfo[] => {
  const excludeIds = proposalIdSet(exclusion);
  return proposals.filter(({ id }) => !excludeIds.has(id as ProposalId));
};

export type ProposalInfoMap = {
  id: ProposalId | undefined;
  proposal: Proposal | undefined;
  proposer: NeuronId | undefined;
  title: string | undefined;
  url: string | undefined;
  color: ProposalStatusColor | undefined;

  created: bigint;
  decided: bigint | undefined;
  executed: bigint | undefined;
  failed: bigint | undefined;
  deadline: bigint | undefined;

  topic: string | undefined;
  topicDescription: string | undefined;
  type: string | undefined;
  typeDescription: string | undefined;
  status: ProposalStatus;
  statusString: string;
  statusDescription: string | undefined;
  rewardStatus: ProposalRewardStatus;
  rewardStatusString: string;
  rewardStatusDescription: string | undefined;
};

export const mapProposalInfo = (
  proposalInfo: ProposalInfo
): ProposalInfoMap => {
  const {
    proposal,
    proposer,
    id,
    status,
    rewardStatus,
    deadlineTimestampSeconds,
    proposalTimestampSeconds,
    decidedTimestampSeconds,
    executedTimestampSeconds,
    failedTimestampSeconds,
  } = proposalInfo;

  const {
    topics,
    topics_description,
    status_description,
    status: statusLabels,
    rewards,
    rewards_description,
  } = get(i18n);
  const deadline =
    deadlineTimestampSeconds === undefined
      ? undefined
      : deadlineTimestampSeconds - BigInt(nowInSeconds());

  const topicKey = Topic[proposalInfo?.topic];

  const statusKey = ProposalStatus[status];
  const rewardStatusKey = ProposalRewardStatus[rewardStatus];

  return {
    id,
    proposer,
    proposal,
    title: proposal?.title,
    url: proposal?.url,
    color: PROPOSAL_COLOR[status],

    created: proposalTimestampSeconds,
    decided: decidedTimestampSeconds > 0 ? decidedTimestampSeconds : undefined,
    executed:
      executedTimestampSeconds > 0 ? executedTimestampSeconds : undefined,
    failed: failedTimestampSeconds > 0 ? failedTimestampSeconds : undefined,
    deadline,

    topic: keyOf({ obj: topics, key: topicKey }),
    topicDescription: keyOf({ obj: topics_description, key: topicKey }),
    status,
    statusString: keyOf({ obj: statusLabels, key: statusKey }),
    statusDescription: keyOf({ obj: status_description, key: statusKey }),
    rewardStatus,
    rewardStatusString: keyOf({ obj: rewards, key: rewardStatusKey }),
    rewardStatusDescription: keyOf({
      obj: rewards_description,
      key: rewardStatusKey,
    }),
    ...mapProposalType(proposal),
  };
};

/**
 * If the action is a ExecuteNnsFunction, then we map the NNS function id (its detailed label).
 * Otherwise, we map the action function itself.
 *
 * This outcome is called "the proposal type".
 */
const mapProposalType = (
  proposal: Proposal | undefined
): Pick<ProposalInfoMap, "type" | "typeDescription"> => {
  const {
    actions,
    actions_description,
    nns_functions,
    nns_functions_description,
  } = get(i18n);

  const NO_MATCH = { type: undefined, typeDescription: undefined };

  if (proposal === undefined) {
    return NO_MATCH;
  }

  const nnsFunctionKey: string | undefined = getNnsFunctionKey(proposal);

  if (nnsFunctionKey !== undefined) {
    return {
      type: keyOf({ obj: nns_functions, key: nnsFunctionKey }),
      typeDescription: keyOf({
        obj: nns_functions_description,
        key: nnsFunctionKey,
      }),
    };
  }

  const action: string | undefined = proposalFirstActionKey(proposal);

  return action !== undefined
    ? {
        type: keyOf({ obj: actions, key: action }),
        typeDescription: keyOf({ obj: actions_description, key: action }),
      }
    : NO_MATCH;
};

/**
 * A proposal can be accepted or declined if the majority votes before its duration expires but, it remains open for voting until then.
 * That is why we should not consider the status "OPEN" to present a proposal as open for voting but consider the duration.
 */
export const isProposalDeadlineInTheFuture = (
  proposalInfo: ProposalInfo
): boolean => votingPeriodEnd(proposalInfo).getTime() >= Date.now();

/**
 * Return the voting period end date of a proposal.
 * @returns {Date} The voting period end date
 */
const votingPeriodEnd = (proposalInfo: ProposalInfo): Date => {
  const { deadlineTimestampSeconds } = proposalInfo;

  // Fallback for undefined deadline_timestamp_seconds which actually should never happen.
  if (deadlineTimestampSeconds === undefined) {
    return votingPeriodEndFallback(proposalInfo);
  }

  return new Date(Number(deadlineTimestampSeconds) * 1000);
};

// Following fallback logic follows what's originally implemented by the Dashboard

const SHORT_VOTING_PERIOD_SECONDS = 60 * 60 * 12; // 12 hours
const WAIT_FOR_QUIET_THRESHOLD_SECONDS = 60 * 60 * 24 * 4; // 4 days

const votingPeriodEndFallback = ({
  proposalTimestampSeconds,
  topic,
}: ProposalInfo): Date => {
  const durationInSeconds = [Topic.ManageNeuron, Topic.ExchangeRate].includes(
    topic
  )
    ? SHORT_VOTING_PERIOD_SECONDS
    : WAIT_FOR_QUIET_THRESHOLD_SECONDS;

  return new Date(
    Number(proposalTimestampSeconds) * 1000 + durationInSeconds * 1000
  );
};

/** Update proposal voting state as it participated in voting */
export const updateProposalVote = ({
  proposalInfo,
  neuron,
  vote,
}: {
  proposalInfo: ProposalInfo;
  neuron: NeuronInfo;
  vote: Vote;
}): ProposalInfo => {
  const { votingPower, neuronId } = neuron;
  const votedBallot: Ballot = {
    neuronId,
    vote,
    votingPower,
  };

  return {
    ...proposalInfo,
    ballots: [
      ...proposalInfo.ballots.filter(
        ({ neuronId }) => neuronId !== neuron.neuronId
      ),
      votedBallot,
    ],
    latestTally: {
      ...(proposalInfo.latestTally as Tally),
      yes:
        vote === Vote.Yes
          ? (proposalInfo.latestTally?.yes ?? BigInt(0)) + votingPower
          : proposalInfo.latestTally?.yes ?? BigInt(0),
      no:
        vote === Vote.No
          ? (proposalInfo.latestTally?.no ?? BigInt(0)) + votingPower
          : proposalInfo.latestTally?.no ?? BigInt(0),
    },
  };
};

/** Returns `registerVote` error reason text or undefined if not an error */
const registerVoteErrorReason = (
  neuronId: NeuronId,
  result: PromiseSettledResult<void>
): string | undefined => {
  if (result.status === "fulfilled") {
    return undefined;
  }

  const reason =
    result.reason instanceof Error ? errorToString(result.reason) : undefined;
  // detail text
  return replacePlaceholders(get(i18n).error.register_vote_neuron, {
    $neuronId: neuronId.toString(),
    $reason:
      reason === undefined || reason?.length === 0
        ? get(i18n).error.fail
        : reason,
  });
};

/** Returns `registerVote` error details (neuronId and the reason by error) */
export const registerVoteErrorDetails = ({
  responses,
  neuronIds,
}: {
  responses: PromiseSettledResult<void>[];
  neuronIds: bigint[];
}): string[] => {
  const details: string[] = responses
    .map((response, i) => registerVoteErrorReason(neuronIds[i], response))
    .filter(isDefined);

  return details;
};

/** There are neurons in a queue whose vote is not yet been registered */
export const voteRegistrationActive = (votes: VoteRegistration[]): boolean =>
  votes.find(
    ({ neuronIds, successfullyVotedNeuronIds }) =>
      neuronIds.length > successfullyVotedNeuronIds.length
  ) !== undefined;
